import json
import datetime
import argparse
import os
import pathlib

import numpy as np
import jax
import jax.numpy as jnp

os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"

import diffrax
import optax
from haiku import PRNGSequence

import exciting_environments as excenvs

from dmpe.utils.signals import aprbs
from dmpe.utils.density_estimation import select_bandwidth, get_uniform_target_distribution
from dmpe.excitation.excitation_utils import soft_penalty
from dmpe.algorithms.algorithms import excite_with_dmpe


# file path setup
REPO_ROOT_PATH = pathlib.Path(__file__).parent.parent.parent.parent.parent
TARGETED_DATA_PATH = REPO_ROOT_PATH / pathlib.Path("data") / pathlib.Path("classical_systems")


def safe_json_dump(obj, fp):
    default = lambda o: f"<<non-serializable: {type(o).__qualname__}>>"
    return json.dump(obj, fp, default=default)


parser = argparse.ArgumentParser(description="Process 'sys_name' to choose the system to experiment on.")
parser.add_argument(
    "sys_name",
    metavar="sys_name",
    type=str,
    help="The name of the environment. Options are ['pendulum', 'fluid_tank', 'cart_pole'].",
)
parser.add_argument("--gpu_id", type=int, default=0, help="GPU id to use.")

args = parser.parse_args()
sys_name = args.sys_name

gpus = jax.devices()
gpu_id = args.gpu_id
jax.config.update("jax_default_device", gpus[args.gpu_id])

### Start experiment parameters #######################################################################################
if sys_name == "pendulum":
    ## Start pendulum experiment parameters

    env_params = dict(batch_size=1, tau=2e-2, max_torque=5, g=9.81, l=1, m=1, env_solver=diffrax.Tsit5())
    env = excenvs.make(
        env_id="Pendulum-v0",
        batch_size=env_params["batch_size"],
        action_normalizations={
            "torque": excenvs.utils.MinMaxNormalization(min=-env_params["max_torque"], max=env_params["max_torque"])
        },
        static_params={"g": env_params["g"], "l": env_params["l"], "m": env_params["m"]},
        solver=env_params["env_solver"],
        tau=env_params["tau"],
    )
    alg_params = dict(
        bandwidth=None,
        n_prediction_steps=20,
        points_per_dim=21,
        grid_extend=1.05,
        excitation_optimizer=optax.adabelief(1e-1),
        n_opt_steps=10,
        start_optimizing=5,
        consider_action_distribution=True,
        penalty_function=None,
        target_distribution=None,
        clip_action=True,
        n_starts=5,
        reuse_proposed_actions=True,
    )

    alg_params["bandwidth"] = float(
        select_bandwidth(
            delta_z=2,
            dim=env.physical_state_dim + env.action_dim,
            n_g=alg_params["points_per_dim"],
            percentage=0.3,
        )
    )

    # overwrite penalty function and target distribution
    alg_params["penalty_function"] = lambda x, u: soft_penalty(a=x, a_max=1, penalty_order=2) + soft_penalty(
        a=u, a_max=1, penalty_order=2
    )
    alg_params["target_distribution"] = get_uniform_target_distribution(
        dim=3 if alg_params["consider_action_distribution"] else 2,
        points_per_dim=alg_params["points_per_dim"],
        bandwidth=alg_params["bandwidth"],
        grid_extend=alg_params["grid_extend"],
        consider_action_distribution=alg_params["consider_action_distribution"],
        penalty_function=alg_params["penalty_function"],
    )

    exp_params = dict(
        seed=None,
        n_time_steps=15_000,
        model_class=None,
        env_params=env_params,
        alg_params=alg_params,
        model_trainer_params=None,
        model_params=None,
    )
    seeds = list(np.arange(101, 131))
    ## End pendulum experiment parameters

elif sys_name == "fluid_tank":
    ## Start fluid_tank experiment parameters

    env_params = dict(
        batch_size=1,
        tau=5,
        max_height=3,
        max_inflow=0.2,
        base_area=jnp.pi,
        orifice_area=jnp.pi * 0.1**2,
        c_d=0.6,
        g=9.81,
        env_solver=diffrax.Tsit5(),
    )
    env = excenvs.make(
        "FluidTank-v0",
        physical_normalizations=dict(height=excenvs.utils.MinMaxNormalization(min=0, max=env_params["max_height"])),
        action_normalizations=dict(inflow=excenvs.utils.MinMaxNormalization(min=0, max=env_params["max_inflow"])),
        static_params=dict(
            base_area=env_params["base_area"],
            orifice_area=env_params["orifice_area"],
            c_d=env_params["c_d"],
            g=env_params["g"],
        ),
        tau=env_params["tau"],
        solver=env_params["env_solver"],
    )

    alg_params = dict(
        bandwidth=None,
        n_prediction_steps=10,
        points_per_dim=50,
        grid_extend=1.05,
        excitation_optimizer=optax.adabelief(1e-1),
        n_opt_steps=10,
        start_optimizing=5,
        consider_action_distribution=True,
        penalty_function=None,
        target_distribution=None,
        clip_action=True,
        n_starts=5,
        reuse_proposed_actions=True,
    )

    alg_params["bandwidth"] = float(
        select_bandwidth(
            delta_z=2,
            dim=env.physical_state_dim + env.action_dim,
            n_g=alg_params["points_per_dim"],
            percentage=0.3,
        )
    )

    # overwrite penalty function and target distribution
    alg_params["penalty_function"] = lambda x, u: soft_penalty(a=x, a_max=1, penalty_order=2) + soft_penalty(
        a=u, a_max=1, penalty_order=2
    )
    alg_params["target_distribution"] = get_uniform_target_distribution(
        dim=2 if alg_params["consider_action_distribution"] else 1,
        points_per_dim=alg_params["points_per_dim"],
        bandwidth=alg_params["bandwidth"],
        grid_extend=alg_params["grid_extend"],
        consider_action_distribution=alg_params["consider_action_distribution"],
        penalty_function=alg_params["penalty_function"],
    )

    exp_params = dict(
        seed=None,
        n_time_steps=15_000,
        model_class=None,
        env_params=env_params,
        alg_params=alg_params,
        model_trainer_params=None,
        model_params=None,
    )
    seeds = list(np.arange(101, 131))
    ## End fluid_tank experiment parameters

elif sys_name == "cart_pole":
    ## Start cart_pole experiment parameters

    env_params = dict(
        batch_size=1,
        tau=2e-2,
        max_force=10,
        static_params={
            "mu_p": 0.002,
            "mu_c": 0.5,
            "l": 0.5,
            "m_p": 0.1,
            "m_c": 1,
            "g": 9.81,
        },
        physical_normalizations={
            "deflection": excenvs.utils.MinMaxNormalization(min=-2.4, max=2.4),
            "velocity": excenvs.utils.MinMaxNormalization(min=-8, max=8),
            "theta": excenvs.utils.MinMaxNormalization(min=-jnp.pi, max=jnp.pi),
            "omega": excenvs.utils.MinMaxNormalization(min=-8, max=8),
        },
        env_solver=diffrax.Tsit5(),
    )
    env = excenvs.make(
        env_id="CartPole-v0",
        batch_size=env_params["batch_size"],
        action_normalizations={
            "force": excenvs.utils.MinMaxNormalization(min=-env_params["max_force"], max=env_params["max_force"])
        },
        physical_normalizations=env_params["physical_normalizations"],
        static_params=env_params["static_params"],
        solver=env_params["env_solver"],
        tau=env_params["tau"],
    )

    alg_params = dict(
        bandwidth=0.12,
        n_prediction_steps=50,
        points_per_dim=10,
        grid_extend=1.05,
        excitation_optimizer=optax.adabelief(1e-1),
        n_opt_steps=5,
        start_optimizing=5,
        consider_action_distribution=True,
        penalty_function=None,
        target_distribution=None,
        clip_action=True,
        n_starts=5,
        reuse_proposed_actions=True,
    )

    # alg_params["bandwidth"] = float(
    #     select_bandwidth(
    #         delta_z=2,
    #         dim=env.physical_state_dim + env.action_dim,
    #         n_g=alg_params["points_per_dim"],
    #         percentage=0.1,
    #     )
    # )

    # overwrite penalty function and target distribution
    alg_params["penalty_function"] = lambda x, u: soft_penalty(a=x, a_max=1, penalty_order=2) + soft_penalty(
        a=u, a_max=1, penalty_order=2
    )
    alg_params["target_distribution"] = get_uniform_target_distribution(
        dim=5 if alg_params["consider_action_distribution"] else 4,
        points_per_dim=alg_params["points_per_dim"],
        bandwidth=alg_params["bandwidth"],
        grid_extend=alg_params["grid_extend"],
        consider_action_distribution=alg_params["consider_action_distribution"],
        penalty_function=alg_params["penalty_function"],
    )

    exp_params = dict(
        seed=None,
        n_time_steps=15_000,
        model_class=None,
        env_params=env_params,
        alg_params=alg_params,
        model_trainer_params=None,
        model_params=None,
    )
    seeds = list(np.arange(101, 131))

    ## End cart_pole experiment parameters

else:
    raise NotImplementedError(f"System '{sys_name}' is unknown. Choose from ['pendulum', 'fluid_tank', 'cart_pole'].")

### End experiment parameters #########################################################################################

for exp_idx, seed in enumerate(seeds):
    print("Running experiment", exp_idx, f"(seed: {seed}) on '{sys_name}'")
    exp_params["seed"] = int(seed)

    # Check that the targeted data folder actually exist:
    results_path = TARGETED_DATA_PATH / pathlib.Path("perfect_model_dmpe") / pathlib.Path(sys_name)
    print(f"Results will be written to: '{results_path}'.")
    assert results_path.exists(), (
        f"The expected results path '{results_path}' does not seem to exist. Please create the necessary file structure "
        + "or adapt the path."
    )

    # setup PRNG
    key = jax.random.PRNGKey(seed=exp_params["seed"])
    data_key, _, _, expl_key, key = jax.random.split(key, 5)
    data_rng = PRNGSequence(data_key)

    # initial guess
    proposed_actions = jnp.hstack(
        [
            aprbs(alg_params["n_prediction_steps"], env.batch_size, 1, 10, next(data_rng))[0]
            for _ in range(env.action_dim)
        ]
    )
    # run excitation algorithm
    observations, actions, model, density_estimate, losses, proposed_actions, _ = excite_with_dmpe(
        env,
        exp_params,
        proposed_actions,
        None,
        expl_key,
    )

    # save parameters
    file_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    with open(results_path / pathlib.Path(f"params_{file_name}.json"), "w") as fp:
        safe_json_dump(exp_params, fp)

    # save observations + actions
    with open(results_path / pathlib.Path(f"data_{file_name}.json"), "w") as fp:
        json.dump(dict(observations=observations.tolist(), actions=actions.tolist()), fp)

    jax.clear_caches()

### End experiments ###################################################################################################



"""A collection of functions for loading and evaluating experiment results."""

from typing import Callable
import json
import pathlib
import glob
import os

import matplotlib.pyplot as plt
import numpy as np
import jax.numpy as jnp

from dmpe.models.model_utils import load_model
from dmpe.evaluation.plotting_utils import plot_sequence, plot_model_performance
from dmpe.evaluation.metrics_utils import default_jsd, default_ae, default_mcudsa, default_ksfc


def get_experiment_ids(results_path: pathlib.Path):
    """Get the experiment ids available in the specified directory."""
    json_file_paths = glob.glob(str(results_path / pathlib.Path("*.json")))
    identifiers = set([pathlib.Path(path).stem.split("_", maxsplit=1)[-1] for path in json_file_paths])
    return sorted(list(identifiers))


def get_organized_experiment_ids(full_results_path, force_consider_actions=False):
    """Only relevant for the PMSM.

    Get the experiment ids available in the specified directory, organized by their 'rpm' and whether
    the action distribution has been considered in the experiment.
    """
    experiment_ids = get_experiment_ids(full_results_path)
    organized_experiment_ids = {}

    for experiment_id in experiment_ids:

        ca = experiment_id.split("ca_")[-1].split("_")[0] == "True"

        if ca not in organized_experiment_ids.keys():
            organized_experiment_ids[ca] = {}

        rpm = float(experiment_id.split("rpm_")[-1].split("_")[0])
        if rpm not in organized_experiment_ids[ca].keys():
            organized_experiment_ids[ca][rpm] = []
        organized_experiment_ids[ca][rpm].append(experiment_id)

    if force_consider_actions:
        if True not in organized_experiment_ids.keys():
            print("No experiments with consider_actions=True. Regard experiments to consider actions.")
            organized_experiment_ids[True] = organized_experiment_ids[False]
            del organized_experiment_ids[False]
        else:
            del organized_experiment_ids[False]

    return organized_experiment_ids


def load_experiment_results(exp_id: str, results_path: pathlib.Path, model_class=None, to_array=True):
    """Load the experiment data for the experiment with identifier exp_id at the results path."""
    if os.path.isfile(results_path / pathlib.Path(f"params_{exp_id}.json")):
        with open(results_path / pathlib.Path(f"params_{exp_id}.json"), "rb") as fp:
            params = json.load(fp)
    else:
        params = None

    with open(results_path / pathlib.Path(f"data_{exp_id}.json"), "rb") as fp:
        data = json.load(fp)

        if to_array:
            try:
                observations = jnp.array(data["observations"])
                actions = jnp.array(data["actions"])
            except:
                try:
                    observations = jnp.array(np.concatenate(data["observations"]))
                    actions = jnp.array(np.concatenate(data["actions"]))
                except:
                    obs_dim = np.array(data["observations"][0]).shape[0]

                    observations = np.stack(data["observations"][:-obs_dim], axis=0)
                    observations = jnp.array(
                        np.concatenate([observations, np.array(data["observations"][-obs_dim:])[None, ...]])
                    )
                    actions = jnp.array(np.stack(data["actions"]))
        else:
            observations = data["observations"]
            actions = data["actions"]

    if model_class is not None:
        model = load_model(results_path / pathlib.Path(f"model_{exp_id}.json"), model_class)
        return params, observations, actions, model
    else:
        return params, observations, actions, None


def evaluate_experiment_metrics(observations, actions, metrics, featurize=None):
    """Evaluate the given observations and actions using the specified metrics."""
    results = {}

    if metrics is None:
        metrics = {
            "jsd": default_jsd,
            "ae": default_ae,
            "mcudsa": default_mcudsa,
            "ksfc": default_ksfc,
        }

    for name, metric in metrics.items():
        results[name] = metric(observations, actions).item()

    if featurize is not None:
        assert isinstance(featurize, Callable)
        for name, metric in metrics.items():
            results[f"{name}_feat"] = metric(featurize(observations), actions).item()

    return results


def evaluate_algorithm_metrics(identifiers, results_path, featurize=None):
    """Iterate over the given experiment identifiers and evaluate the corresponding observations and actions."""
    results = {}
    for identifier in identifiers:
        _, observations, actions, _ = load_experiment_results(
            exp_id=identifier, results_path=results_path, model_class=None
        )
        single_result = evaluate_experiment_metrics(observations, actions, featurize=featurize)

        if len(results.keys()) == 0:
            for key, value in single_result.items():
                results[key] = [value]
        else:
            for key, value in single_result.items():
                results[key].append(value)
    return results


def extract_metrics_over_timesteps(experiment_ids, results_path, lengths, metrics=None, slotted=False):
    """Iterate over the given experiment identifiers and evaluate the corresponding observations and actions.
    Only the first 'length' elements are considered for each 'length' in the 'lengths'-list.
    """
    all_results = []
    for idx, identifier in enumerate(experiment_ids):
        print(f"Experiment {identifier} at index {idx}")

        _, observations, actions, _ = load_experiment_results(
            exp_id=identifier,
            results_path=results_path,
            model_class=None,
        )
        if slotted:
            single_results = []

            for i in range(len(lengths) - 1):
                single_results.append(
                    evaluate_experiment_metrics(
                        observations[lengths[i] : lengths[i + 1]], actions[lengths[i] : lengths[i + 1]], metrics=metrics
                    )
                )

        else:
            single_results = [
                evaluate_experiment_metrics(observations[:N], actions[:N], metrics=metrics) for N in lengths
            ]
        metric_keys = single_results[0].keys()

        results_by_metric = {key: [] for key in metric_keys}
        for result in single_results:
            for metric_key in metric_keys:
                results_by_metric[metric_key].append(result[metric_key])

        all_results.append(results_by_metric)

    print("Reshape to results by metric...")
    all_results_by_metric = {key: [] for key in metric_keys}
    for result in all_results:
        for metric_key in metric_keys:
            all_results_by_metric[metric_key].append(result[metric_key])

    for metric_key in all_results_by_metric.keys():
        all_results_by_metric[metric_key] = jnp.stack(jnp.array(all_results_by_metric[metric_key]))

    print("Done")

    return all_results_by_metric


def extract_metrics_over_timesteps_via_interpolation(experiment_ids, results_path, target_lengths, metrics=None):
    all_raw_results = []
    all_raw_lengths = []

    all_results = []
    for idx, identifier in enumerate(experiment_ids):
        print(f"Experiment {identifier} at index {idx} via interpolation")

        _, observations, actions, _ = load_experiment_results(
            exp_id=identifier,
            results_path=results_path,
            model_class=None,
            to_array=False,
        )

        raw_lengths = [len(subsequence) for subsequence in observations]
        raw_lengths = np.cumsum(raw_lengths[:-1])

        observations = jnp.array(np.concatenate(observations))
        actions = jnp.array(np.concatenate(actions))
        raw_results = [evaluate_experiment_metrics(observations[:N], actions[:N], metrics=metrics) for N in raw_lengths]

        metric_keys = raw_results[0].keys()

        raw_results_by_metric = {key: [] for key in metric_keys}
        for result in raw_results:
            for metric_key in metric_keys:
                raw_results_by_metric[metric_key].append(result[metric_key])

        all_raw_results.append(raw_results_by_metric)
        all_raw_lengths.append(raw_lengths)

        interpolated_results_by_metric = {}
        for metric_key in metric_keys:
            interpolated_results_by_metric[metric_key] = jnp.interp(
                x=target_lengths,
                xp=raw_lengths,
                fp=jnp.array(raw_results_by_metric[metric_key]),
            )
        all_results.append(interpolated_results_by_metric)

    print("Reshape to results by metric...")
    all_results_by_metric = {key: [] for key in metric_keys}
    for result in all_results:
        for metric_key in metric_keys:
            all_results_by_metric[metric_key].append(result[metric_key])

    for metric_key in all_results_by_metric.keys():
        all_results_by_metric[metric_key] = jnp.stack(jnp.array(all_results_by_metric[metric_key]))

    print("Done")

    return {"interp": all_results_by_metric, "raw": all_raw_results, "raw_lengths": all_raw_lengths}


def quick_eval(env, identifier, results_path, model_class=None):
    """Gives a quick overview over the specified experiment."""

    params, observations, actions, model = load_experiment_results(
        exp_id=identifier, results_path=results_path, model_class=model_class
    )

    print(identifier)
    print(params["alg_params"])

    if observations.shape[0] == actions.shape[0]:
        actions = actions[:-1]

    fig, axs = plot_sequence(
        observations=observations,
        actions=actions,
        tau=env.tau,
        obs_labels=env.obs_description,
        action_labels=env.action_description,
    )
    plt.show()

    if model is not None:
        fig, axs = plot_model_performance(
            model=model,
            true_observations=observations[:1000],
            actions=actions[:999],
            tau=env.tau,
            obs_labels=env.obs_description,
            action_labels=[r"$u$"],
        )
        plt.show()



import json
import datetime
import argparse
import pathlib

import numpy as np
import jax
import jax.numpy as jnp
import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"

jax.config.update("jax_platform_name", "cpu")

import diffrax

import exciting_environments as excenvs

from dmpe.excitation.excitation_utils import soft_penalty
from dmpe.related_work.algorithms import excite_with_iGOATS


# file path setup
REPO_ROOT_PATH = pathlib.Path(__file__).parent.parent.parent.parent.parent
TARGETED_DATA_PATH = REPO_ROOT_PATH / pathlib.Path("data") / pathlib.Path("classical_systems")


def safe_json_dump(obj, fp):
    default = lambda o: f"<<non-serializable: {type(o).__qualname__}>>"
    return json.dump(obj, fp, default=default)


parser = argparse.ArgumentParser(description="Process 'sys_name' to choose the system to experiment on.")
parser.add_argument(
    "sys_name",
    metavar="sys_name",
    type=str,
    help="The name of the environment. Options are ['pendulum', 'fluid_tank', 'cart_pole'].",
)

args = parser.parse_args()
sys_name = args.sys_name


### Start experiment parameters #######################################################################################

if sys_name == "pendulum":
    ## Start pendulum experiment parameters

    env_params = dict(batch_size=1, tau=2e-2, max_torque=5, g=9.81, l=1, m=1, env_solver=diffrax.Tsit5())
    env = excenvs.make(
        env_id="Pendulum-v0",
        batch_size=env_params["batch_size"],
        action_normalizations={
            "torque": excenvs.utils.MinMaxNormalization(min=-env_params["max_torque"], max=env_params["max_torque"])
        },
        static_params={"g": env_params["g"], "l": env_params["l"], "m": env_params["m"]},
        solver=env_params["env_solver"],
        tau=env_params["tau"],
    )

    h = 10
    a = 10

    penalty_function = lambda x, u: 1e3 * soft_penalty(a=x, a_max=1, penalty_order=2) + 1e3 * soft_penalty(
        a=u, a_max=1, penalty_order=2
    )

    alg_params = dict(
        prediction_horizon=h,
        application_horizon=a,
        bounds_amplitude=(-1, 1),
        bounds_duration=(10, 100),
        population_size=50,
        n_generations=25,
        featurize=lambda x: x,
        rng=None,
        compress_data=True,
        compression_target_N=500,
        compression_feat_dim=-2,
        compression_dist_th=0.1,
        penalty_order=penalty_function,
    )
    seeds = list(np.arange(101, 131))
    ## End pendulum experiment parameters

elif sys_name == "fluid_tank":
    ## Start pendulum experiment parameters

    env_params = dict(
        batch_size=1,
        tau=5,
        max_height=3,
        max_inflow=0.2,
        base_area=jnp.pi,
        orifice_area=jnp.pi * 0.1**2,
        c_d=0.6,
        g=9.81,
        env_solver=diffrax.Tsit5(),
    )
    env = excenvs.make(
        "FluidTank-v0",
        physical_normalizations=dict(height=excenvs.utils.MinMaxNormalization(min=0, max=env_params["max_height"])),
        action_normalizations=dict(inflow=excenvs.utils.MinMaxNormalization(min=0, max=env_params["max_inflow"])),
        static_params=dict(
            base_area=env_params["base_area"],
            orifice_area=env_params["orifice_area"],
            c_d=env_params["c_d"],
            g=env_params["g"],
        ),
        tau=env_params["tau"],
        solver=env_params["env_solver"],
    )

    h = 10
    a = 10

    penalty_function = lambda x, u: 1e3 * soft_penalty(a=x, a_max=1, penalty_order=2) + 1e3 * soft_penalty(
        a=u, a_max=1, penalty_order=2
    )

    alg_params = dict(
        prediction_horizon=h,
        application_horizon=a,
        bounds_amplitude=(-1, 1),
        bounds_duration=(5, 50),
        population_size=50,
        n_generations=25,
        featurize=lambda x: x,
        rng=None,
        compress_data=True,
        compression_target_N=500,
        compression_feat_dim=-2,
        compression_dist_th=0.1,
        penalty_function=penalty_function,
    )

    seeds = list(np.arange(101, 131))
    ## End fluid_tank experiment parameters

elif sys_name == "cart_pole":
    ## Start cart_pole experiment parameters

    env_params = dict(
        batch_size=1,
        tau=2e-2,
        max_force=10,
        static_params={
            "mu_p": 0.002,
            "mu_c": 0.5,
            "l": 0.5,
            "m_p": 0.1,
            "m_c": 1,
            "g": 9.81,
        },
        physical_normalizations={
            "deflection": excenvs.utils.MinMaxNormalization(min=-2.4, max=2.4),
            "velocity": excenvs.utils.MinMaxNormalization(min=-8, max=8),
            "theta": excenvs.utils.MinMaxNormalization(min=-jnp.pi, max=jnp.pi),
            "omega": excenvs.utils.MinMaxNormalization(min=-8, max=8),
        },
        env_solver=diffrax.Tsit5(),
    )
    env = excenvs.make(
        env_id="CartPole-v0",
        batch_size=env_params["batch_size"],
        action_normalizations={
            "force": excenvs.utils.MinMaxNormalization(min=-env_params["max_force"], max=env_params["max_force"])
        },
        physical_normalizations=env_params["physical_normalizations"],
        static_params=env_params["static_params"],
        solver=env_params["env_solver"],
        tau=env_params["tau"],
    )

    h = 10
    a = 5  # to help with stabilization?

    penalty_function = lambda x, u: 1e3 * soft_penalty(a=x, a_max=1, penalty_order=2) + 1e3 * soft_penalty(
        a=u, a_max=1, penalty_order=2
    )

    alg_params = dict(
        prediction_horizon=h,
        application_horizon=a,
        bounds_amplitude=(-1, 1),
        bounds_duration=(1, 100),
        population_size=50,
        n_generations=25,
        featurize=lambda x: x,
        rng=None,
        compress_data=True,
        compression_target_N=500,
        compression_feat_dim=-2,
        compression_dist_th=0.1,
        penalty_function=penalty_function,
    )

    seeds = list(np.arange(101, 131))

    ## End cart_pole experiment parameters

### End experiment parameters #########################################################################################


### Start experiments #################################################################################################

for exp_idx, seed in enumerate(seeds):

    print("Running experiment", exp_idx, f"(seed: {seed}) on '{sys_name}'")

    # Check that the targeted data folder actually exist:
    results_path = TARGETED_DATA_PATH / pathlib.Path("igoats") / pathlib.Path(sys_name)
    print(f"Results will be written to: '{results_path}'.")
    assert results_path.exists(), (
        f"The expected results path '{results_path}' does not seem to exist. Please create the necessary file structure "
        + "or adapt the path."
    )

    exp_params = dict(
        n_time_steps=15_000,
        seed=int(seed),
        alg_params=alg_params,
        env_params=env_params,
    )

    # run excitation algorithm
    observations, actions = excite_with_iGOATS(
        n_time_steps=exp_params["n_time_steps"],
        env=env,
        prediction_horizon=alg_params["prediction_horizon"],
        application_horizon=alg_params["application_horizon"],
        bounds_amplitude=alg_params["bounds_amplitude"],
        bounds_duration=alg_params["bounds_duration"],
        population_size=alg_params["population_size"],
        n_generations=alg_params["n_generations"],
        featurize=alg_params["featurize"],
        rng=np.random.default_rng(seed),
        compress_data=alg_params["compress_data"],
        compression_target_N=alg_params["compression_target_N"],
        compression_feat_dim=alg_params["compression_feat_dim"],
        compression_dist_th=alg_params["compression_dist_th"],
        penalty_function=alg_params["penalty_function"],
        plot_subsequences=False,
    )

    observations = [obs.tolist() for obs in observations]
    actions = [act.tolist() for act in actions]

    # save parameters
    file_name = datetime.datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
    with open(results_path / pathlib.Path(f"params_{file_name}.json"), "w") as fp:
        safe_json_dump(exp_params, fp)

    # save observations + actions
    with open(results_path / pathlib.Path(f"data_{file_name}.json"), "w") as fp:
        json.dump(dict(observations=observations, actions=actions), fp)

    jax.clear_caches()

### End experiments ###################################################################################################



import numpy as np


def simulate_ahead_with_env(env, obs, state, actions):
    observations = []
    observations.append(obs)

    for i in range(actions.shape[0]):
        obs, state = env.step(state, actions[i, :], env.env_properties)
        observations.append(obs)

    return np.vstack(observations), state
